@@ -14,7 +14,7 @@
 
 /**
  * @defgroup radcli-api Main API
- * @brief Main API Functions 
+ * @brief Main API Functions
  *
  * @{
  */
@@ -26,6 +26,13 @@
 #include "util.h"
 #include "tls.h"
 
+#ifndef TRUE
+#define TRUE  1
+#define FALSE 0
+#endif
+
+static int rc_conf_int_2(rc_handle const *rh, char const *optname, int complain);
+
 /** Find an option in the option list
  *
  * @param rh a handle to parsed configuration.
@@ -121,104 +128,108 @@ static int set_option_srv(char const *filename, int line, OPTION *option, char c
 		serv->max = 0;
 	}
 
-	if (serv->max > 0) {
-		DEBUG(LOG_ERR, "cannot set multiple servers");
-		return -1;
-	}
-
 	p_pointer = strtok_r(p_dupe, ", \t", &p_save);
 
-	/* check to see for '[IPv6]:port' syntax */
-	if ((q = strchr(p_pointer,'[')) != NULL) {
-		*q = '\0';
-		q++;
-		p_pointer = q;
+        while(p_pointer != NULL) {
+                if (serv->max > SERVER_MAX) {
+                        DEBUG(LOG_ERR, "cannot set more than %d servers", SERVER_MAX);
+                        return -1;
+                }
 
-		q = strchr(p_pointer, ']');
-		if (q == NULL) {
-			free(p_dupe);
-			rc_log(LOG_CRIT, "read_config: IPv6 parse error");
-			return -1;
-		}
-		*q = '\0';
-		q++;
+		DEBUG(LOG_ERR, "processing server: %s", p_pointer);
+                /* check to see for '[IPv6]:port' syntax */
+                if ((q = strchr(p_pointer,'[')) != NULL) {
+                        *q = '\0';
+                        q++;
+                        p_pointer = q;
 
-		if (q[0] == ':') {
-			q++;
-		}
+                        q = strchr(p_pointer, ']');
+                        if (q == NULL) {
+                                free(p_dupe);
+                                rc_log(LOG_CRIT, "read_config: IPv6 parse error");
+                                return -1;
+                        }
+                        *q = '\0';
+                        q++;
 
-		/* Check to see if we have '[IPv6]:port:secret' syntax */
-		if((s=strchr(q, ':')) != NULL) {
-			*s = '\0';
-			s++;
-			serv->secret[serv->max] = strdup(s);
-			if (serv->secret[serv->max] == NULL) {
-				rc_log(LOG_CRIT, "read_config: out of memory");
-				if (option->val == NULL) {
-					free(p_dupe);
-					free(serv);
-				}
-				return -1;
-			}
-		}
+                        if (q[0] == ':') {
+                                q++;
+                        }
 
-	} else /* Check to see if we have 'servername:port' syntax */
-	if ((q = strchr(p_pointer,':')) != NULL) {
-		*q = '\0';
-		q++;
+                        /* Check to see if we have '[IPv6]:port:secret' syntax */
+                        if((s=strchr(q, ':')) != NULL) {
+                                *s = '\0';
+                                s++;
+                                serv->secret[serv->max] = strdup(s);
+                                if (serv->secret[serv->max] == NULL) {
+                                        rc_log(LOG_CRIT, "read_config: out of memory");
+                                        if (option->val == NULL) {
+                                                free(p_dupe);
+                                                free(serv);
+                                        }
+                                        return -1;
+                                }
+                        }
 
-		/* Check to see if we have 'servername:port:secret' syntax */
-		if((s = strchr(q,':')) != NULL) {
-			*s = '\0';
-			s++;
-			serv->secret[serv->max] = strdup(s);
-			if (serv->secret[serv->max] == NULL) {
-				rc_log(LOG_CRIT, "read_config: out of memory");
-				if (option->val == NULL) {
-					free(p_dupe);
-					free(serv);
-				}
-				return -1;
-			}
-		}
-	}
+                } else /* Check to see if we have 'servername:port' syntax */
+                        if ((q = strchr(p_pointer,':')) != NULL) {
+                                *q = '\0';
+                                q++;
 
-	if(q && strlen(q) > 0) {
-		serv->port[serv->max] = atoi(q);
-	} else {
-		if (!strcmp(option->name,"authserver"))
-			if ((svp = getservbyname ("radius", "udp")) == NULL)
-				serv->port[serv->max] = PW_AUTH_UDP_PORT;
-			else
-				serv->port[serv->max] = ntohs ((unsigned int) svp->s_port);
-		else if (!strcmp(option->name, "acctserver"))
-			if ((svp = getservbyname ("radacct", "udp")) == NULL)
-				serv->port[serv->max] = PW_ACCT_UDP_PORT;
-			else
-				serv->port[serv->max] = ntohs ((unsigned int) svp->s_port);
-		else {
-			rc_log(LOG_ERR, "%s: line %d: no default port for %s", filename, line, option->name);
-			if (option->val == NULL) {
-				free(p_dupe);
-				free(serv);
-			}
-			return -1;
-		}
-	}
+                                /* Check to see if we have 'servername:port:secret' syntax */
+                                if((s = strchr(q,':')) != NULL) {
+                                        *s = '\0';
+                                        s++;
+                                        serv->secret[serv->max] = strdup(s);
+                                        if (serv->secret[serv->max] == NULL) {
+                                                rc_log(LOG_CRIT, "read_config: out of memory");
+                                                if (option->val == NULL) {
+                                                        free(p_dupe);
+                                                        free(serv);
+                                                }
+                                                return -1;
+                                        }
+                                }
+                        }
 
-	serv->name[serv->max] = strdup(p_pointer);
-	if (serv->name[serv->max] == NULL) {
-		rc_log(LOG_CRIT, "read_config: out of memory");
-		if (option->val == NULL) {
-			free(p_dupe);
-			free(serv);
-		}
-		return -1;
-	}
-	free(p_dupe);
+                if(q && strlen(q) > 0) {
+                        serv->port[serv->max] = atoi(q);
+                } else {
+                        if (!strcmp(option->name,"authserver"))
+                                if ((svp = getservbyname ("radius", "udp")) == NULL)
+                                        serv->port[serv->max] = PW_AUTH_UDP_PORT;
+                                else
+                                        serv->port[serv->max] = ntohs ((unsigned int) svp->s_port);
+                        else if (!strcmp(option->name, "acctserver"))
+                                if ((svp = getservbyname ("radacct", "udp")) == NULL)
+                                        serv->port[serv->max] = PW_ACCT_UDP_PORT;
+                                else
+                                        serv->port[serv->max] = ntohs ((unsigned int) svp->s_port);
+                        else {
+                                rc_log(LOG_ERR, "%s: line %d: no default port for %s", filename, line, option->name);
+                                if (option->val == NULL) {
+                                        free(p_dupe);
+                                        free(serv);
+                                }
+                                return -1;
+                        }
+                }
 
-	serv->max++;
+                serv->name[serv->max] = strdup(p_pointer);
+                if (serv->name[serv->max] == NULL) {
+                        rc_log(LOG_CRIT, "read_config: out of memory");
+                        if (option->val == NULL) {
+                                free(p_dupe);
+                                free(serv);
+                        }
+                        return -1;
+                }
 
+                serv->max++;
+                p_pointer = strtok_r(NULL, ", \t", &p_save);
+        }
+
+        free(p_dupe);
 	if (option->val == NULL)
 		option->val = (void *)serv;
 
@@ -389,6 +400,106 @@ rc_handle *rc_config_init(rc_handle *rh)
 	return rh;
 }
 
+static ssize_t plain_sendto(void *ptr, int sockfd,
+			    const void *buf, size_t len, int flags,
+			    const struct sockaddr *dest_addr, socklen_t addrlen)
+{
+	return sendto(sockfd, buf, len, flags, dest_addr, addrlen);
+}
+
+static ssize_t plain_tcp_sendto(void *ptr, int sockfd,
+			    const void *buf, size_t len, int flags,
+			    const struct sockaddr *dest_addr, socklen_t addrlen)
+{
+	if((connect(sockfd, dest_addr, addrlen)) != 0){
+		rc_log(LOG_ERR, "%s: Connect Call Failed : %s", __FUNCTION__, strerror(errno));
+		return -1;
+	}
+	return sendto(sockfd, buf, len, flags, dest_addr, addrlen);
+}
+
+static ssize_t plain_recvfrom(void *ptr, int sockfd,
+			      void *buf, size_t len, int flags,
+			      struct sockaddr *src_addr, socklen_t * addrlen)
+{
+	return recvfrom(sockfd, buf, len, flags, src_addr, addrlen);
+}
+
+static void plain_close_fd(int fd)
+{
+	close(fd);
+}
+
+static int plain_get_fd(void *ptr, struct sockaddr *our_sockaddr)
+{
+	int sockfd;
+
+	sockfd = socket(our_sockaddr->sa_family, SOCK_DGRAM, 0);
+	if (sockfd < 0) {
+		return -1;
+	}
+
+	if (our_sockaddr->sa_family == AF_INET)
+		((struct sockaddr_in *)our_sockaddr)->sin_port = 0;
+	else
+		((struct sockaddr_in6 *)our_sockaddr)->sin6_port = 0;
+
+	if (bind(sockfd, SA(our_sockaddr), SA_LEN(our_sockaddr)) < 0) {
+		close(sockfd);
+		return -1;
+	}
+	return sockfd;
+}
+
+static int plain_tcp_get_fd(void *ptr, struct sockaddr *our_sockaddr)
+{
+	int sockfd;
+
+	sockfd = socket(our_sockaddr->sa_family, SOCK_STREAM, 0);
+	if (sockfd < 0) {
+		return -1;
+	}
+
+	if (our_sockaddr->sa_family == AF_INET)
+		((struct sockaddr_in *)our_sockaddr)->sin_port = 0;
+	else
+		((struct sockaddr_in6 *)our_sockaddr)->sin6_port = 0;
+
+	if (bind(sockfd, SA(our_sockaddr), SA_LEN(our_sockaddr)) < 0) {
+		close(sockfd);
+		return -1;
+	}
+	return sockfd;
+}
+
+const static rc_sockets_override default_socket_funcs = {
+	.get_fd = plain_get_fd,
+	.close_fd = plain_close_fd,
+	.sendto = plain_sendto,
+	.recvfrom = plain_recvfrom
+};
+
+const static rc_sockets_override default_tcp_socket_funcs = {
+	.get_fd = plain_tcp_get_fd,
+	.close_fd = plain_close_fd,
+	.sendto = plain_tcp_sendto,
+	.recvfrom = plain_recvfrom
+};
+
+static int set_addr(struct sockaddr_storage *ss, const char *ip)
+{
+	memset(ss, 0, sizeof(*ss));
+	if (inet_pton(AF_INET, ip, &((struct sockaddr_in *)ss)->sin_addr) == 1) {
+		ss->ss_family = AF_INET;
+	} else if (inet_pton(AF_INET6, ip, &((struct sockaddr_in6 *)ss)->sin6_addr) == 1) {
+		ss->ss_family = AF_INET6;
+	} else {
+		rc_log(LOG_CRIT, "invalid IP address for nas-ip: %s", ip);
+		return -1;
+	}
+	return 0;
+}
+
 static int apply_config(rc_handle *rh)
 {
 	const char *txt;
@@ -398,33 +509,55 @@ static int apply_config(rc_handle *rh)
 	rh->own_bind_addr_set = 0;
 	rc_own_bind_addr(rh, &rh->own_bind_addr);
 	rh->own_bind_addr_set = 1;
-#ifdef HAVE_GNUTLS
-	txt = rc_conf_str(rh, "serv-auth-type");
+
+	txt = rc_conf_str(rh, "nas-ip");
 	if (txt != NULL) {
-		if (strcasecmp(txt, "dtls") == 0) {
-		    	ret = rc_init_tls(rh, SEC_FLAG_DTLS);
-		} else if (strcasecmp(txt, "tls") == 0) {
-		    	ret = rc_init_tls(rh, 0);
-		} else {
-			rc_log(LOG_CRIT, "unknown server authentication type: %s", txt);
+		if (set_addr(&rh->nas_addr, txt) < 0)
 			return -1;
-		}
-
-	    	if (ret < 0) {
-	    		rc_log(LOG_CRIT, "error initializing %s", txt);
-			return -1;
-		}
+		rh->nas_addr_set = 1;
 	}
-#endif
 
-	return 0;	
+	txt = rc_conf_str(rh, "serv-type");
+	if (txt == NULL)
+		txt = rc_conf_str(rh, "serv-auth-type");
+
+	if (txt == NULL)
+		txt = "udp";
+
+	if (strcasecmp(txt, "udp") == 0) {
+		memset(&rh->so, 0, sizeof(rh->so));
+		rh->so_type = RC_SOCKET_UDP;
+		memcpy(&rh->so, &default_socket_funcs, sizeof(rh->so));
+		ret = 0;
+	} else if (strcasecmp(txt, "tcp") == 0) {
+		memset(&rh->so, 0, sizeof(rh->so));
+		rh->so_type = RC_SOCKET_TCP;
+		memcpy(&rh->so, &default_tcp_socket_funcs, sizeof(rh->so));
+		ret = 0;
+#ifdef HAVE_GNUTLS
+	} else if (strcasecmp(txt, "dtls") == 0) {
+		ret = rc_init_tls(rh, SEC_FLAG_DTLS);
+	} else if (strcasecmp(txt, "tls") == 0) {
+		ret = rc_init_tls(rh, 0);
+#endif
+	} else {
+		rc_log(LOG_CRIT, "unknown server type: %s", txt);
+		return -1;
+	}
+
+	if (ret < 0) {
+		rc_log(LOG_CRIT, "error initializing %s", txt);
+		return -1;
+	}
+
+	return 0;
 
 }
 
 /** Read the global config file
  *
- * This function will load the provided configuration file, and 
- * any other files such as the dictionary. 
+ * This function will load the provided configuration file, and
+ * any other files such as the dictionary.
  *
  * Note: To preserve compatibility with libraries of the same API
  * which don't load the dictionary care is taken not to reload the
@@ -547,6 +680,13 @@ rc_handle *rc_read_config(char const *filename)
 		return NULL;
 	}
 
+        {
+                int clientdebug = rc_conf_int_2(rh, "clientdebug", FALSE);
+                if(clientdebug > 0) {
+                        radcli_debug = clientdebug;
+                }
+        }
+
 	p = rc_conf_str(rh, "dictionary");
 	if (p != NULL) {
 		if (rc_read_dictionary(rh, p) != 0) {
@@ -581,13 +721,13 @@ char *rc_conf_str(rc_handle const *rh, char const *optname)
 	}
 }
 
-/** Get the value of a config option
+/*- Get the value of a config option
  *
  * @param rh a handle to parsed configuration.
  * @param optname the name of an option.
  * @return config option value.
  */
-int rc_conf_int(rc_handle const *rh, char const *optname)
+static int rc_conf_int_2(rc_handle const *rh, char const *optname, int complain)
 {
 	OPTION *option;
 
@@ -596,16 +736,21 @@ int rc_conf_int(rc_handle const *rh, char const *optname)
 	if (option != NULL) {
 		if (option->val) {
 			return *((int *)option->val);
-		} else {
+		} else if(complain) {
 			rc_log(LOG_ERR, "rc_conf_int: config option %s was not set", optname);
-			return 0;
 		}
+                return 0;
 	} else {
 		rc_log(LOG_CRIT, "rc_conf_int: unkown config option requested: %s", optname);
 		return 0;
 	}
 }
 
+int rc_conf_int(rc_handle const *rh, char const *optname)
+{
+        return rc_conf_int_2(rh, optname, TRUE);
+}
+
 /** Get the value of a config option
  *
  * @param rh a handle to parsed configuration.
@@ -692,8 +837,8 @@ static int find_match (const struct addrinfo* addr, const struct addrinfo *hostn
 			len1 = SA_GET_INLEN(ptr->ai_addr);
 			len2 = SA_GET_INLEN(ptr2->ai_addr);
 
-			if (len1 > 0 && 
-			    len1 == len2 && 
+			if (len1 > 0 &&
+			    len1 == len2 &&
 			    memcmp(SA_GET_INADDR(ptr->ai_addr), SA_GET_INADDR(ptr2->ai_addr), len1) == 0) {
 				return 0;
 			}
@@ -712,7 +857,7 @@ static int find_match (const struct addrinfo* addr, const struct addrinfo *hostn
 static int rc_ipaddr_local(const struct sockaddr *addr)
 {
 	int temp_sock, res, serrno;
-	struct sockaddr tmpaddr;
+	struct sockaddr_storage tmpaddr;
 
 	memcpy(&tmpaddr, addr, SA_LEN(addr));
 
@@ -725,7 +870,7 @@ static int rc_ipaddr_local(const struct sockaddr *addr)
 	} else {
 		((struct sockaddr_in6*)&tmpaddr)->sin6_port = 0;
 	}
-	res = bind(temp_sock, &tmpaddr, SA_LEN(&tmpaddr));
+	res = bind(temp_sock, SA(&tmpaddr), SS_LEN(&tmpaddr));
 	serrno = errno;
 	close(temp_sock);
 	if (res == 0)
@@ -763,7 +908,7 @@ static int rc_is_myname(const struct addrinfo *info)
  * @param info: will hold a pointer to addrinfo
  * @param secret will hold the server's secret (of %MAX_SECRET_LENGTH).
  * @param type %AUTH or %ACCT
- 
+
  * @return 0 on success, -1 on failure.
  */
 int rc_find_server_addr (rc_handle const *rh, char const *server_name,
@@ -891,7 +1036,7 @@ int rc_find_server_addr (rc_handle const *rh, char const *server_name,
 			 server_name, rc_conf_str(rh, "servers"));
 		goto fail;
 	}
-	
+
 	result = 0;
 	goto cleanup;
 
@@ -910,7 +1055,10 @@ int rc_find_server_addr (rc_handle const *rh, char const *server_name,
  * rc_config_free:
  * @param rh a handle to parsed configuration
  *
- * Free allocated config values
+ * Free allocated config values. For legacy compatibility
+ * reasons this will not release any dictionary entries.
+ * To release all memory from the handle use rc_destroy()
+ * instead.
  *
  */
 void rc_config_free(rc_handle *rh)
@@ -979,3 +1127,9 @@ rc_socket_type rc_get_socket_type(rc_handle *rh)
 }
 
 /** @} */
+ /*
+ * Local Variables:
+ * c-basic-offset:8
+ * c-style: whitesmith
+ * End:
+ */
